/*
 *  Created on Oct 14, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.bot2.commlib;

import org.nukesoft.bot2.core.*;
import org.nukesoft.nukeutils.*;

/**
 * The CommPort class wraps an externaly loaded CommLib and handles such tasks
 * as connection maintnance and error correction.
 * 
 * @author Ryan Michela
 */
public class CommPort extends Thread implements CommLibCallback
{
	private String loggin;
	private String password;
	private String libPath;
	private String bindName;
	private int pingDelay;
	private boolean keepGoing = true;
	private boolean initialized;
	private CommLib comm;
	private CommPortCallback callback;
	private Log log;
	
	/**
	 * @param loggin The loggin for this port to use.
	 * @param password The associated password.
	 * @param libPath The path to the communications library to load.
	 * @param pingDelay The number of seconds between pings. Zero for infinite.
	 * @param callback Class implementing the CommPortCallback interface
	 * used to receive incomming message events.
	 * @param bindName The bind name of this CommPort
	 */
	public CommPort(String loggin, String password, String libPath, int pingDelay, CommPortCallback callback, String bindName)
	{
		super(bindName);
		
		this.loggin = loggin;
		this.password = password;
		this.libPath = libPath;
		this.pingDelay = pingDelay;
		this.callback = callback;
		this.bindName = bindName;
		this.log = new Log(bindName + ".commlog");
		
		AbsolutePathClassLoader apcl = new AbsolutePathClassLoader();
		initialized = false;
		try
		{
			//load the CommLib
			comm = (CommLib)apcl.loadClass(libPath).newInstance();
			comm.bindCallback(this);
			
			if(comm.connect(loggin, password) == true)
			{
				//Start the maintnance thread
				this.start();
				VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
										"CommPort" +
										VerboseConsole.GRAY_TEXT() +
										": Connected to " +  
										VerboseConsole.BOLD_ON() + VerboseConsole.BROWN_TEXT() +
										loggin + ":" + comm.identifyProtocal() +
										VerboseConsole.BOLD_OFF() + VerboseConsole.GRAY_TEXT(), 
										VerboseConsole.HIGH);
				log.logEntry("Connection established to " + loggin + ":" + comm.identifyProtocal());
				initialized = true;
			}
			else
			{
				VerboseConsole.println(VerboseConsole.RED_TEXT() +
										"CommPort" +
										VerboseConsole.GRAY_TEXT() +
										": Failed to connect to " + loggin + ":" + comm.identifyProtocal(), 
										VerboseConsole.CRITICAL);
				log.logEntry("Failed to connect to " + loggin + ":" + comm.identifyProtocal());
			}
		}
		catch (Exception e)
		{
			VerboseConsole.println(VerboseConsole.RED_TEXT() +
									"CommPort" +
									VerboseConsole.GRAY_TEXT() +
									": CommLib load failed! " + libPath, 
									VerboseConsole.CRITICAL);	
			log.logEntry("CommLib load failed!");
		} 
	}
	
	/**
	 * Shuts down the CommPort and disconnects from the IM network.
	 */
	public void shutdown()
	{
		if(comm != null)
		{	
			VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
									"CommPort" +
									VerboseConsole.GRAY_TEXT() +
									": Shutting down connection " + 
									VerboseConsole.BOLD_ON() + VerboseConsole.BROWN_TEXT() +
									loggin + ":" + comm.identifyProtocal() +
									VerboseConsole.BOLD_OFF() + VerboseConsole.GRAY_TEXT(), 
									VerboseConsole.HIGH);
			log.logEntry("Shutting down connection " + loggin + ":" + comm.identifyProtocal());
	
			//disconnect from the IM network
			comm.disconnect();
		}
		//stop the maintnance thread
		keepGoing = false;
		this.interrupt();
	}
	
	/**
	 * Sends an instant message over the network. Breaks the message into
	 * 1024 character chunks and applies a message throttleing algorithm
	 * to the transmition.
	 * @param message The message to send.
	 */
	public void sendMessage(IM message)
	{
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"CommPort" +
								VerboseConsole.GRAY_TEXT() +
								": Attempting to send message over " + loggin + ":" + comm.identifyProtocal(), 
								VerboseConsole.LOW);
		
		int BUFFSIZE = 1024;
		int cursor = 0;
		int spaceOffset;
		String buf;
		String sourceString = message.message;
		
		//chop the string source string up into BUFFSIZE length parts
		while(cursor + BUFFSIZE <= sourceString.length())
		{	
			buf = sourceString.substring(cursor, cursor + BUFFSIZE);
			spaceOffset = buf.length() - buf.lastIndexOf(' ');
			buf = sourceString.substring(cursor, cursor + BUFFSIZE - spaceOffset);
			
			if(cursor == 0)
			{
				comm.sendMessage(new IM(message.screenName, "=== Starting Big Block Transmition ==="));
			}
			comm.sendMessage(new IM(message.screenName, buf));
			
			cursor = cursor + BUFFSIZE - spaceOffset;
			
			//delay longer as the number of sequential posts increases to avoid
			//overloading server speed limit
			try
			{
				if(cursor < BUFFSIZE * 2)
				{
					Thread.sleep(500);
				}
				else if(cursor < BUFFSIZE * 4)
				{
					Thread.sleep(2000);
				}
				else if(cursor < BUFFSIZE * 6)
				{
					Thread.sleep(4000);
				}
				else if(cursor < BUFFSIZE * 8)
				{
					Thread.sleep(6000);
				}
				else
				{
					Thread.sleep(10000);
				}
			}
			catch(Exception e)
			{
			
			}
		}
		if(cursor + BUFFSIZE > sourceString.length())
		{	
			buf = sourceString.substring(cursor);
			comm.sendMessage(new IM(message.screenName, buf));
			if(cursor != 0)
			{
				comm.sendMessage(new IM(message.screenName, "=== Ending Big Block Transmition =="));
			}
		}
	}
	
	/**
	 * Attempts to warn an individual. May not have an effect on all networks.
	 * @param screenName The target.
	 * @param anonymous If true, attempt to warn anonymously.
	 */	
	public void sendWarning(String screenName, boolean anonymous)
	{
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"CommPort" +
								VerboseConsole.GRAY_TEXT() +
								": Attempting to send warning over " + loggin + ":" + comm.identifyProtocal(), 
								VerboseConsole.LOW);
		comm.sendWarning(screenName, anonymous);
	}
	
	/**
	 * Handles in incomming error.
	 * @see org.nukesoft.bot2.commlib.CommLibCallback#handleError(java.lang.String)
	 */
	public void handleError(String error)
	{
		VerboseConsole.println(VerboseConsole.RED_TEXT() +
								"CommPort" +
								VerboseConsole.GRAY_TEXT() +
								": An error has occured on " + loggin + ":" + comm.identifyProtocal(), 
								VerboseConsole.CRITICAL);
		log.logEntry("Error! " + error);
	}

	/**
	 * Handles an incomming instant message.
	 * @see org.nukesoft.bot2.commlib.CommLibCallback#handleIM(org.nukesoft.bot2.core.IM)
	 */
	public void handleIM(IM message)
	{
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"CommPort" +
								VerboseConsole.GRAY_TEXT() +
								": Incomming IM over " + loggin + ":" + comm.identifyProtocal(), 
								VerboseConsole.LOW);
		callback.handleIM(message, this);
	}

	/**
	 * Handles incomming warnings.
	 * @param screenName The screen name that did the warning.
	 * @see org.nukesoft.bot2.commlib.CommLibCallback#handleWarn(java.lang.String)
	 */
	public void handleWarn(String screenName)
	{
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"CommPort" +
								VerboseConsole.GRAY_TEXT() +
								": Incomming warning over " + loggin + ":" + comm.identifyProtocal(), 
								VerboseConsole.LOW);
		log.logEntry("Received warning. Level now at " + getWarningLevel() + "%");
		callback.handleWarn(screenName, this);
	}
	
	/**
	 * Returns the CommLib warning level.
	 * @see org.nukesoft.bot2.commlib.CommPort#getWarningLevel()
	 * @return Warning level.
	 */
	public String getWarningLevel()
	{
		return comm.getWarningLevel();
	}
	
	/**
	 * Gets the Bind Name for theis CommPort.
	 * @return The bind name.
	 */
	public String getBindName()
	{
		return bindName;
	}
	
	/**
	 * Main loop for the connection maintnance thread. Pings the connection and waits
	 * for a response.
	 * @see java.lang.Runnable#run()
	 */
	public void run()
	{
		//keep looping until told to stop
		if(pingDelay != 0)
		{	
			while(keepGoing)
			{
				try
				{
					sleep(pingDelay * 1000);
					VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
											"CommPort" +
											VerboseConsole.GRAY_TEXT() +
											": Initiating ping on " + loggin + ":" + comm.identifyProtocal(), 
											VerboseConsole.HIGH);
					log.logEntry("Sending ping.");
					comm.doPing();
				}
				catch(InterruptedException e)
				{
					//break.
				}
			}
		}
	}
}
