/*
 *  Created on Oct 29, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.bot2.core;

import org.nukesoft.bot2.commlib.*;
import org.nukesoft.bot2.lineproc.*;

import org.nukesoft.nukeutils.*;

import java.util.*;

/**
 * A Dialog is the bridge between CommPort objects and LineProc objects.
 * They serve to provide the internal glue that binds the two ends togeather.
 * All Dialog objects are managed by the DialogManager class.
 * 
 * @author Ryan Michela
 */
public class Dialog extends Observable implements LineProcContextCallback, Observer
{
	private DialogManager manager;
	private CommPort comm;
	private DynamicLoadLPContext lineProc;
	private String identity;
	private int age;
	private Log log;
	private String adminEscapeSequence;
	private String adminLineProcBindID;
	private MessageThread mt = null;
	
	/**
	 * Creates a dialog binding.
	 */
	public Dialog(DialogManager manager, 
				  CommPort comm, 
				  DynamicLoadLPContext lineProc, 
				  String identity,
				  String adminEscapeSequence,
				  String adminLineProcBindID)
	{
		this.manager = manager;
		this.comm = comm;
		this.lineProc = lineProc;
		this.identity = identity;
		this.adminEscapeSequence = adminEscapeSequence;
		this.adminLineProcBindID = adminLineProcBindID;
		age = 0;
		log = new Log(identity + ".botlog");
		mt = new MessageThread();
	}

	/**
	 * Handles incomming message from the CommLib.
	 * @param message The message to foreward.
	 */
	public void handleIM(IM message)
	{
		age = 0;
		log.logEntry(message.screenName + "\t\t" + message.message);
		VerboseConsole.println(VerboseConsole.CYAN_TEXT() + 
								identity + 
								VerboseConsole.BOLD_ON() + VerboseConsole.BROWN_TEXT() +
								" <<< " + 
								VerboseConsole.BOLD_OFF() + VerboseConsole.MAGENTA_TEXT() +
								message.message +
								VerboseConsole.GRAY_TEXT(), 
								VerboseConsole.MEDIUM);
		
		//handle conversation bind redirection
		String redirect = identity + " :: (rec) " + message.message;
		this.setChanged();
		this.notifyObservers(redirect);
		
		//handle admin escape loading
		if(!adminEscapeSequence.equals("") && message.message.equals(adminEscapeSequence))
		{
			if(lineProc.loadNewLineProc(adminLineProcBindID))
			{
				log.logEntry("*** Entering Admin Mode ***");
				return;
			}
			else
			{
				comm.sendMessage(new IM(getRemoteScreenName(), "Failed to load administrative line processor."));
			}
		}
		
		mt.handleIM(message);
	}

	/**
	 * Handles incomming warning from the CommLib.
	 * @param screenName The screen name doing the warning.
	 */
	public void handleWarn(String screenName)
	{
		log.logEntry(screenName + "\t\t***WARN***");
		VerboseConsole.println(VerboseConsole.CYAN_TEXT() +
								identity + 
								VerboseConsole.BOLD_ON() + VerboseConsole.BROWN_TEXT() +
								" <<< " +
								VerboseConsole.BOLD_OFF() + VerboseConsole.RED_TEXT() +
								"***WARN***" +
								VerboseConsole.GRAY_TEXT(), 
								VerboseConsole.MEDIUM);
		
		//handle conversation bind redirection
		String redirect = identity + " :: (rec) ***WARN***";
		this.setChanged();
		this.notifyObservers(redirect);
		
		mt.handleWarn(screenName);
	}

	/**
	 * Returns a list of active conversations.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#getConversations()
	 */
	public Set getConversations()
	{
		return manager.getConversations();
	}

	/**
	 * Returns this dialog's identity.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#getIdentity()
	 */
	public String getIdentity()
	{
		return identity;
	}

	/**
	 * Gets the warning level from the CommLib.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#getWarningLevel()
	 */
	public String getWarningLevel()
	{
		return comm.getWarningLevel();
	}

	/**
	 * Sends an outbound IM.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#sendIM(org.nukesoft.bot2.core.IM)
	 */
	public void sendIM(IM message)
	{
		age = 0;
		log.logEntry(comm.getName() + "\t\t" + message.message);
		VerboseConsole.println(VerboseConsole.CYAN_TEXT() +
								identity + 
								VerboseConsole.BOLD_ON() + VerboseConsole.BROWN_TEXT() +
								" >>> " +
								VerboseConsole.BOLD_OFF() + VerboseConsole.MAGENTA_TEXT() + 
								message.message + 
								VerboseConsole.GRAY_TEXT(), 
								VerboseConsole.MEDIUM);
		
		//handle conversation bind redirection
		String redirect = identity + " :: (snd) " + message.message;
		this.setChanged();
		this.notifyObservers(redirect);
		
		comm.sendMessage(message);
	}

	/**
	 * Sends an outbound warning.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#sendWarning(java.lang.String, boolean)
	 */
	public void sendWarning(String screenName, boolean anonymous)
	{
		//handle conversation bind redirection
		String redirect = identity + " :: (rec) ***WARN***";
		VerboseConsole.println(VerboseConsole.CYAN_TEXT() +
								identity + 
								VerboseConsole.BOLD_ON() + VerboseConsole.BROWN_TEXT() +
								" >>> " +
								VerboseConsole.BOLD_OFF() + VerboseConsole.RED_TEXT() +
								"***WARN***" +
								VerboseConsole.GRAY_TEXT(), 
								VerboseConsole.MEDIUM);
		this.setChanged();
		this.notifyObservers(redirect);
		
		if(anonymous)
		{
			log.logEntry(comm.getName() + "/t/t***ANONYMOUS WARN***");
		}
		else
		{
			log.logEntry(comm.getName() + "/t/t***WARN***");
		}
		comm.sendWarning(screenName, anonymous);
	}
	
	/**
	 * Binds this dialog to another.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#bindConversation(java.lang.String)
	 */
	public boolean bindConversation(String dialogID)
	{
		Dialog peer = manager.dialogLookup(dialogID);
		if(peer != null)
		{	
			peer.addObserver(this);
			VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
									"Dialog" +
									VerboseConsole.GRAY_TEXT() +
									": " + identity + " ===> " + dialogID, 
									VerboseConsole.LOW);
			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * Unbinds this dialog from another.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#unbindConversation(java.lang.String)
	 */
	public boolean unbindConversation(String dialogID)
	{
		Dialog peer = manager.dialogLookup(dialogID);
		if(peer != null)
		{	
			peer.deleteObserver(this);
			VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
									"Dialog" +
									VerboseConsole.GRAY_TEXT() +
									": " + identity + " =x=> " + dialogID, 
									VerboseConsole.LOW);
			return true;
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * Incraments the age by one.
	 */
	public void tickAge()
	{
		age++;
	}
	
	/**
	 * Gets the Dialog age.
	 * @return Age in seconds.
	 */
	public int getAge()
	{
		return age;
	}
	
	/**
	 * Cleanly shuts down the Dialog.
	 */
	public void shutdown()
	{
		//release all observers
		this.deleteObservers();
		
		//release self from all observer links
		Iterator it = manager.getConversations().iterator();
		while(it.hasNext())
		{
			String peer = (String)it.next();
			unbindConversation(peer);
		}
		
		comm = null;
		lineProc.shutdown();
		lineProc = null;
		mt.shutdown();
		log.close();
	}
	/**
	 * Handles incomming message forewards.
	 * @param arg The third party IM
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable o, Object arg)
	{
		//Foreward incomming redirected message.
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"Dialog" +
								VerboseConsole.GRAY_TEXT() +
								": Redirecting message to " + getRemoteScreenName(), 
								VerboseConsole.LOW);
		comm.sendMessage(new IM(getRemoteScreenName(), (String) arg));
	}
	
	/**
	 * Utility method used for extracting the remote screen name from the
	 * dialog's identity.
	 * @return The remote screen name.
	 */
	private String getRemoteScreenName()
	{
		return identity.substring(identity.lastIndexOf('~')+1);
	}

	/**
	 * @see java.util.Observable#addObserver(java.util.Observer)
	 */
	public synchronized void addObserver(Observer o)
	{
		super.addObserver(o);
		log.logEntry("Adding observer: " + ((Dialog)o).getIdentity());
	}

	/**
	 * @see java.util.Observable#deleteObserver(java.util.Observer)
	 */
	public synchronized void deleteObserver(Observer o)
	{
		super.deleteObserver(o);
		log.logEntry("Removing observer: " + ((Dialog)o).getIdentity());
	}

	/**
	 * Returns a set of comm port bind ids.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#getComPortBindIDs()
	 */
	public Set getComPortBindIDs()
	{
		return manager.getCommPortBindIDs();
	}
	/**
	 * Returns a set of line proc bind ids.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#getLineProcBindIDs()
	 */
	public Set getLineProcBindIDs()
	{
		return manager.getLineProcBindIDs();
	}
	/**
	 * Initiates a conversation.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#initiateConversation(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void initiateConversation(
		String commPortBindID,
		String remoteScreenName,
		String salutation)
	{
		CommPort cp = manager.commPortLookup(commPortBindID);
		IM message = new IM(remoteScreenName, salutation);
		if(cp != null)
		{
			//manager.handleIM(message, cp);
			Dialog foreward = manager.injectNewDialog(cp, remoteScreenName);
			foreward.sendIM(message);
		}
	}

	/**
	 * Interjects text into a conversation. This is the sending half of the interjection
	 * system. When a dialog recieves a message to interject, it obtains a reference to
	 * the target Dialog and calls the handleInterjection() method on it.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#interject(java.lang.String, java.lang.String)
	 */
	public void interject(String dialogID, String interjection)
	{
		Dialog peer = manager.dialogLookup(dialogID);
		if(peer != null)
		{
			peer.handleInterjection(interjection);
		}
	}
	
	/**
	 * Handles an incomming interjection. This is receiving half of the interjection system.
	 */
	void handleInterjection(String interjection)
	{
		comm.sendMessage(new IM(getRemoteScreenName(), interjection));
	}

	/**
	 * Returns the line proc lib path based on bind ID.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#getLineProcLibPath(java.lang.String)
	 */
	public String getLineProcLibPath(String lineProcBindID)
	{
		return manager.lineProcLookup(lineProcBindID);
	}

	/**
	 * Tells a peer to load a new line proc.
	 * @see org.nukesoft.bot2.lineproc.LineProcContextCallback#loadNewLineProcOnPeer(java.lang.String, java.lang.String)
	 */
	public boolean loadNewLineProcOnPeer(String dialogID, String lineProcBindID)
	{
		Dialog peer = manager.dialogLookup(dialogID);
		if(peer != null)
		{
			return peer.handleLoadNewLineProcOnPeer(lineProcBindID);
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * Handles an incomming line proc load request.
	 */
	boolean handleLoadNewLineProcOnPeer(String lineProcBindID)
	{
		return lineProc.loadNewLineProc(lineProcBindID);
	}
	
	/**
	 * The MessageThread class is an internal class of the Dialog class. It
	 * is responsible for grabbing incomming messages and warnings and
	 * forewarding them on down the call chain. The reason for doing this in
	 * another thread is to allow the line processor to incorporate delays
	 * without tying up the event thread in the comm libs. Without this,
	 * messages could not process in parallel, and under high loads, some messages
	 * could be dropped.
	 * @author Ryan Michela
	 */
	private class MessageThread extends Thread
	{
		Object q;
		Object lock = new Object();
		boolean buisy = false;
		boolean keepGoing = true;
		
		public MessageThread()
		{
			super(getIdentity() + " message thread");
			this.start();
		}
		
		/**
		 * Handles an incomming IM.
		 * @param message The IM to handle.
		 */
		public void handleIM(IM message)
		{
			if(!buisy)
			{
				q = message;
				synchronized(lock)
				{
					lock.notify();
				}
			}
		}
		
		/**
		 * Handles an incommng warning.
		 * @param screenName The screen name that sent the warning.
		 */
		public void handleWarn(String screenName)
		{
			if(!buisy)
			{
				q = screenName;
				synchronized(lock)
				{
					lock.notify();
				}
			}
		}
		
		/**
		 * Shuts down this thread.
		 */
		public void shutdown()
		{
			keepGoing = false;
			synchronized(lock)
			{
				lock.notify();
			}
		}
		
		public void run()
		{
			while(keepGoing)
			{	
				buisy = false;
				try{synchronized(lock){lock.wait();}}catch(Exception e){}
				
				buisy = true;
				if(keepGoing)
				{	
					if(q instanceof IM)
					{
						Dialog.this.lineProc.handleIM((IM)q);
					}
					else
					{
						Dialog.this.lineProc.handleWarn((String)q);
					}
				}
			}
		}
	}
}
