/*
 *  Created on Oct 29, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.bot2.core;

import org.nukesoft.bot2.commlib.*;
import org.nukesoft.bot2.lineproc.*;
import org.nukesoft.nukeutils.*;

import java.util.*;

/**
 * The DialogManager handles distribution of incomming instant messages,
 * creating a new Dialog object if necesary. It is also responsible for
 * automatic destruction of Dialog objects after a set durration of
 * inactivity.
 * 
 * @author Ryan Michela
 */
public class DialogManager extends Thread implements CommPortCallback
{
	/**
	 * The map of all active Dialogs. Binds fully qualified dialog
	 * handles to Dialog objects.
	 */
	private Map dialogs = new HashMap();
	/**
	 * Binds CommPort objects to their bind name.
	 */
	private Map commPorts = new HashMap();
	/**
	 * Binds LineProc lib paths to their bind name.
	 */
	private Map lineProcs = new HashMap();
	/**
	 * Binds CommPort bind names to LineProc bind names.
	 */
	private Map defaultBinds = new HashMap();
	
	private String adminEscapeSequence;
	private String adminLineProcBindID;
	
	private boolean keepGoing = true;
	private int timeOut;
	
	/**
	 * Builds the DialogManager. For now, this is a hack.
	 * @param timeOut The number of seconds before a Dialog times out.
	 */
	public DialogManager(DMConfig conf)
	{
		super("Dialog Manager");
		
		//Set system delays.
		timeOut = conf.dialogTimeout;
		Log.setTimeBetweenFlushes(conf.logFlushDelay);
		Log.setWorkingDirectory(conf.logWorkingDir);
		
		//Store administrative escape info
		adminEscapeSequence = conf.adminEscapeSequence;
		adminLineProcBindID = conf.adminLineProcBindID;
		
		//Establish LineProcs
		Iterator it = conf.lpc.iterator();
		while(it.hasNext())
		{
			DMConfig.LPConfig lpc = (DMConfig.LPConfig) it.next();
			lineProcs.put(lpc.bindName, lpc.libPath);
		}
		
		//Establish Default Binds
		it = conf.dbc.iterator();
		while(it.hasNext())
		{
			DMConfig.DBConfig dbc = (DMConfig.DBConfig) it.next();
			defaultBinds.put(dbc.commPortBindName, dbc.lineProcBindName);
		}
		
		//Establish and start the CommPorts
		it = conf.cpc.iterator();
		while(it.hasNext())
		{
			DMConfig.CPConfig cpc = (DMConfig.CPConfig) it.next();
			commPorts.put(cpc.bindName, new CommPort(cpc.loggin, cpc.password, 
							cpc.libPath, cpc.pingDelay, this, cpc.bindName));
		}
		
		//Start the maintnance thread
		this.start();
	}

	/**
	 * Handles an inbound IM from the CommLib.
	 * @see org.nukesoft.bot2.commlib.CommPortCallback#handleIM(org.nukesoft.bot2.core.IM, org.nukesoft.bot2.commlib.CommPort)
	 */
	public void handleIM(IM message, CommPort cp)
	{
		String dialogName = cp.getBindName() + "~" + message.screenName;
		
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"Core" +
								VerboseConsole.GRAY_TEXT() +
								": Incomming message for " + dialogName, 
								VerboseConsole.LOW);
		
		synchronized(dialogs)
		{
			if (dialogs.containsKey(dialogName))
			{
				VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
										"Core" +
										VerboseConsole.GRAY_TEXT() +
										": Forewarding message...", 
										VerboseConsole.LOW);
			
				//foreward the message
				Dialog foreward = (Dialog)dialogs.get(dialogName);
				foreward.handleIM(message);
			}
			else
			{
				VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
										"Core" +
										VerboseConsole.GRAY_TEXT() +
										": " + dialogName + " not found. Building new dialog.", 
										VerboseConsole.HIGH);
				VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
										"Core" +
										VerboseConsole.GRAY_TEXT() +
										": Forewarding message...", 
										VerboseConsole.LOW);

				//foreward the message
				Dialog foreward = injectNewDialog(cp, message.screenName);
				foreward.handleIM(message);
			}
		}
	}
	
	/**
	 * Injects a Dialog into the list of active dialogs and performs the associated bindings.
	 * @param cp The CommPort of the incomming connection.
	 * @param dialogName The identity of the new Dialog.
	 * @return The newly created Dialog.
	 */
	Dialog injectNewDialog(CommPort cp, String foreignScreenName)
	{
		  String dialogName = cp.getBindName() + "~" + foreignScreenName;
		
		  //build a new LineProc
		  String lpBindName = (String)defaultBinds.get(cp.getBindName());
		  String lpLibPath = (String)lineProcs.get(lpBindName);
		  DynamicLoadLPContext lp = new DynamicLoadLPContext(lpLibPath);
	
		  //build a new Dialog
		  Dialog foreward = new Dialog(this, cp, lp, dialogName, adminEscapeSequence, adminLineProcBindID);
		  lp.bindCallback(foreward);
		  lp.loadNewLineProc(lpBindName);
	
		  //store the new Dialog in the dialogs hash
		  dialogs.put(dialogName, foreward);
		  
		  return foreward;
	}

	/**
	 * Handles an inbound warning from the CommPort.
	 * @see org.nukesoft.bot2.commlib.CommPortCallback#handleWarn(java.lang.String, org.nukesoft.bot2.commlib.CommPort)
	 */
	public void handleWarn(String screenName, CommPort cp)
	{	
		String dialogName = cp.getName() + "~" + screenName;
		
		VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
								"Core" +
								VerboseConsole.GRAY_TEXT() +
								": Incomming warning for " + screenName, 
								VerboseConsole.LOW);
		
		if (dialogs.containsKey(dialogName))
		{
			VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
									"Core" +
									VerboseConsole.GRAY_TEXT() +
									": Forewarding warning...", 
									VerboseConsole.LOW);

			//foreward the warning
			Dialog foreward = (Dialog)dialogs.get(dialogName);
			foreward.handleWarn(screenName);
		}
		else
		{
			VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
									"Core" +
									VerboseConsole.GRAY_TEXT() +
									": Warned without prior commnuication! Thats odd.", 
									VerboseConsole.LOW);
		}
	}
	
	/**
	 * Shuts down all Dialog, LineProc, and CommPort objects.
	 */
	public void shutdown()
	{
		VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
								"Core" +
								VerboseConsole.GRAY_TEXT() +
								": Initating master shutdown.", 
								VerboseConsole.HIGH);
		//iterate through all the open CommPorts and shut them down
		Iterator it = commPorts.values().iterator();
		while(it.hasNext())
		{
			((CommPort)it.next()).shutdown();
		}
		
		//kill the maintnance thread.
		keepGoing = false;
		this.interrupt();
		
		//purge all logs
		Log.shutdown();
	}
	
	/**
	 * Returns the set of all active conversations.
	 * Used by the Dialog class.
	 * @return A set of strings with fully qualified conversation names.
	 */
	Set getConversations()
	{
		return dialogs.keySet();
	}
	
	/**
	 * Returns a set of Strings containing the bind IDs of all
	 * comports.
	 * @return A set of Strings.
	 */
	Set getCommPortBindIDs()
	{
		return commPorts.keySet();
	}
	
	/**
	 * Returns a set of Strings containing the bind IDs of all
	 * line procs.
	 * @return A set of Strings.
	 */
	Set getLineProcBindIDs()
	{
		return lineProcs.keySet();
	}
	
	/**
	 * Returns a CommPort based on the bind ID. Used by the Dialog class for initiating
	 * conversations.
	 * @param commPortBindID The bind ID to look up.
	 * @return The associated CommPort. Null if no match.
	 */
	CommPort commPortLookup(String commPortBindID)
	{
		return (CommPort)commPorts.get(commPortBindID);
	}
	
	String lineProcLookup(String lineProcBindID)
	{
		return (String)lineProcs.get(lineProcBindID);
	}
	
	/**
	 * Performs a lookup on the dialogs map. Used by the Dialog class.
	 * @param name The name to look up.
	 * @return The associated Dialog object.
	 */
	Dialog dialogLookup(String name)
	{
		return (Dialog)dialogs.get(name);
	}

	/**
	 * Dialog maintnance thread. Handles automatic destructuin of Dialog objects
	 * after a period of inactivity.
	 * @see java.lang.Thread#start()
	 */
	public synchronized void run()
	{
		while(keepGoing)
		{
			try
			{
				//wait one second
				sleep(1000);
				
				synchronized(dialogs)
				{
					//check all Dialogs to see if they have expired.
					Iterator it = dialogs.values().iterator();
					while(it.hasNext())
					{
						Dialog d = (Dialog)it.next();
						if(d.getAge() > timeOut)
						{
							VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
													"Core" +
													VerboseConsole.GRAY_TEXT() +
													": Purging expired dialog.", 
													VerboseConsole.HIGH);
							d.shutdown();
							it.remove();
						}
						else
						{
							d.tickAge();
						}
					}
				}
			}
			catch(InterruptedException e)
			{
				//do nothing
			}
		}
		
		//shutting down. Kill all Dialogs
		synchronized(dialogs)
		{
			Iterator it = dialogs.values().iterator();
			while(it.hasNext())
			{
				Dialog d = (Dialog)it.next();
				d.shutdown();
				it.remove();
			}
		}
	}

}
