import org.nukesoft.bot2.core.IM;
import org.nukesoft.bot2.lineproc.GenericLineProc;

/*
 *  Created on Nov 24, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

import org.nukesoft.nukeutils.*;
import java.io.*;

/**
 * The ConsoleLineProc provides the ability access a remote console via the
 * bot core. Use this line processor with care as it has the potential to be
 * misused. As a security procaution, this line proc requires a password to
 * load and will lock itself out after three failed attempts in the same
 * dialog session. It can only be unlocked by setting the "locked" tag in the
 * config file to FALSE.
 * 
 * @author Ryan Michela
 */
public class ConsoleLineProc extends GenericLineProc implements Runnable
{
	private String shell;
	private String password = "";
	private boolean locked = true;
	private boolean open = false;
	
	private int failCount = 0;
	private Process p = null;
	private boolean keepGoing = true;
	private Thread scraper = new Thread(this, "Console Scraper");
	private Log accessLog = new Log("Console.accesslog");
	
	/**
	 * Loads the line processor and sends the initial greeting.
	 * @see org.nukesoft.bot2.lineproc.LineProc#init(java.lang.String)
	 */
	public void init(String screenName)
	{
		super.init(screenName);
		
		//read the configuration
		try
		{
			XmlConfManager conf = new XmlConfManager("conf/ConsoleConf.xml");
			shell = conf.getOneConfString("/Console/Shell");
			password = conf.getOneConfString("/Console/Password");
			locked = conf.getOneConfString("/Console/Locked").equalsIgnoreCase("TRUE");
		}
		catch(IOException e)
		{
			context.sendIM(new IM(screenName, "Error in XML configuration."));
		}
		
		//display the greeting
		if(!password.equals(""))
		{
			context.sendIM(new IM(screenName, "=== System Console ===<br>Please enter password:"));
		}
		else
		{
			context.sendIM(new IM(screenName, "=== System Console ==="));
		}
	}
	
	/**
	 * Handles the console
	 * @see org.nukesoft.bot2.lineproc.LineProc#handleIM(org.nukesoft.bot2.core.IM)
	 */
	public void handleIM(IM message)
	{
		//the console is locked. kick the user out
		if(locked)
		{
			context.sendIM(new IM(message.screenName, "Console Locked."));
			context.loadNewLineProc(context.lastLoadedLineProc());
			return;
		}
		
		//password not yet confirmed
		if(!open)
		{
			if(failCount < 2)
			{
				if(message.message.equals(password) || password.equals(""))
				{
					context.sendIM(new IM(message.screenName, "Console unlocked."));
					try
					{
						p = Runtime.getRuntime().exec(shell);
						open = true;
						scraper.start();
						accessLog.logEntry("Login: " + context.getIdentity());
						context.sendIM(new IM(message.screenName, "=== " + shell + " ==="));
					}
					catch(IOException e)
					{
						context.sendIM(new IM(message.screenName, "Error starting shell!"));
					}
				}
				else
				{
					failCount++;
					accessLog.logEntry("Invalid Password (try " + (failCount + 1) + "): " + context.getIdentity());
					context.sendIM(new IM(message.screenName, "Invalid password.<br>Enter password:"));
				}
			}
			else
			{
				lockConfig();
				accessLog.logEntry("***Locked due to three consecutive login failures!***: " + context.getIdentity());
				context.sendIM(new IM(message.screenName, "Invalid password.<br>Console locked."));
				context.loadNewLineProc(context.lastLoadedLineProc());
			}
		}
		//password good
		else
		{
			//write the console
			PrintWriter pw = new PrintWriter(new OutputStreamWriter(p.getOutputStream()));
			pw.println(message.message);
			pw.flush();	
		}
	}

	/**
	 * Kills the console process, if it has been created.
	 * @see org.nukesoft.bot2.lineproc.LineProc#shutdown()
	 */
	public void shutdown()
	{
		keepGoing = false;
		scraper.interrupt();
		if(p != null)
		{
			p.destroy();
		}
	}

	/**
	 * Locks the console by locking the config file.
	 * TODO: Implement.
	 */
	private void lockConfig()
	{
		try
		{		
			PrintWriter pw = new PrintWriter(new FileWriter("conf/ConsoleConf.xml"));
			pw.println("<?xml version=\"1.0\"?>");
			pw.println("<!-- Configuration for ConsoleLineProc -->");
			pw.println();
			pw.println("<Console>");
			pw.println("	<Shell>");
			pw.println("		" + shell);
			pw.println("	</Shell>");
			pw.println("	<Password>");
			pw.println("		" + password);
			pw.println("	</Password>");
			pw.println("	<Locked>");
			pw.println("		TRUE");
			pw.println("	</Locked>");
			pw.println("</Console> ");
			pw.close();
		}
		catch(IOException e)
		{
			VerboseConsole.println(VerboseConsole.RED_TEXT() +
									"ConsoleLineProc" +
									VerboseConsole.GRAY_TEXT() +
									": Error locking config file! Possible security hole." + e.getMessage(), 
									VerboseConsole.CRITICAL);
		}
	}
	
	/**
	 * Scrapes the console screen returning data as it comes.
	 * @see java.lang.Runnable#run()
	 */
	public void run()
	{
		InputStream in = p.getInputStream();
		InputStream err = p.getErrorStream();
		int c;
		int e;
		StringBuffer buf = new StringBuffer();
		
		try
		{
			while(keepGoing)
			{
				if(in.available() > 0 || err.available() > 0)
				{	
					//while there are bytes to read on stdin or stderr
					while(in.available() > 0 || err.available() > 0)
					{
						//read a byte from stdin if applicable
						if(in.available() > 0)
						{	
							c = in.read();
						
							if((char)c == '\n')
							{
								buf.append("<br>");
							}
							else if((char)c == '\r')
							{
								//ignore
							}
							else
							{
								buf.append((char)c);
							}
						}
						//read a byte from stderr if applicable
						if(err.available() > 0)
						{	
							e = err.read();
						
							if((char)e == '\n')
							{
								buf.append("<br>");
							}
							else if((char)e == '\r')
							{
								//ignore
							}
							else
							{
								buf.append((char)e);
							}
						}
					}
					//Send the contents of the output buffer
					context.sendIM(new IM(super.remoteScreenName, buf.toString()));
					buf = new StringBuffer();
				}
				else
				{
					Thread.sleep(200);
				}
			}
		}
		catch(Exception error)
		{
			
		}
	}
}
