import org.nukesoft.bot2.core.IM;
import org.nukesoft.bot2.lineproc.GenericLineProc;
import org.nukesoft.nukeutils.*;

import java.net.*;
import org.txml.*;

/*
 *  Created on Nov 12, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/**
 * This class is demonstration of the various things that LineProcs are
 * capeable of. This LineProc accepts a zip code and returns a weather
 * report. The report is based off of the XML web service provided by
 * <a href="http://www.innergears.com">http://www.innergears.com</a>.
 * 
 * <pre>
 * This web services accepts one parameters:
 * 1) A string containing the Zip Code to find the weather for.
 * 
 * This web service returns an array of strings, containing the following elements:
 * [0] = State Code, such as "FL".
 * [1] = City Name, such as "Melbourne".
 * [2] = Latitude of this post office.
 * [3] = Longitude of this post office.
 * [4] = The ICAO Code of the nearest weather reporting station, such as "KLMB".
 * [5] = The distance in miles from the post office to the weather reporting station, such as "12.56".
 * [6] = The direction to the weather reporting station, such as "NE".
 * [7] = The name of the weather reporting station, such as "Melbourne, Melbourne International Airport".
 * [8] = The state code in which the weather reporting station is located, such as "FL".
 * [9] = The country in which the weather reporting station is located, such as "United States".
 * [10] = The latitude of the weather reporting station.
 * [11] = The longitude of the weather reporting station.
 * [12] = The report title, such as "Melbourne, Melbourne International Airport, FL, United States".
 * [13] = The time of the last report, such as "Oct 13, 2002 - 02:53 PM EDT".
 * [14] = The wind direction and speed, such as "from the E (090 degrees) at 12 MPH (10 KT)".
 * [15] = The visibility, such as "10 mile(s)".
 * [16] = The sky condition, such as "partly cloudy".
 * [17] = The Temperature, such as "84.9 F (29.4 C)".
 * [18] = The Dew Point, such as "73.0 F (22.8 C)".
 * [19] = The humidity, such as "67%".
 * [20] = The air pressure, such as "30.03 in. Hg (1016 hPa)".
 * [21] = The entire METAR line received from NOAA, such as "KMLB 131853Z AUTO 09010KT 10SM VCTS SCT020 29/23 A3003 RMK AO2 LTG DSNT S-NW TSB16E36B41 SLP167 T02940228".
 * </pre>
 * 
 * @author Ryan Michela
 */
public class WeatherReportLineProc extends GenericLineProc
{	
	/**
	 * Handles an incomming IM.
	 * @see org.nukesoft.bot2.lineproc.LineProc#handleIM(org.nukesoft.bot2.core.IM)
	 */
	public void handleIM(IM message)
	{
		try
		{
			String zip = message.message;
			if(isZip(zip))
			{	
				context.sendIM(new IM(message.screenName, generateReport(zip)));
			}
			else
			{
				String prompt = "<b>Welcome to the Weather Bot!</b><br>" +
								"Please enter the zip code you would like<br>" +
								"a weather report for:";
				context.sendIM(new IM(message.screenName, prompt));
			}
		}
		catch(Exception e)
		{
			context.sendIM(new IM(message.screenName, "Hmm. Somthing went wrong. Try again later."));
		}
	}
	
	/**
	 * Checks to see if the input string is in zip code format.
	 * @param zip The string to check.
	 * @return TRUE if in proper format. FALSE if not.
	 */
	private boolean isZip(String zip)
	{
		//check the string length
		if(zip.length() != 5)
		{
			return false;
		}
		
		//make sure every character is an integer
		char[] zipChars = zip.toCharArray();
		for(int i = 0; i < 5; i++)
		{
			if(!Character.isDigit(zipChars[i]))
			{
				return false;
			}
		}
		
		return true;
	}
	/**
	 * Generates a weather report by quering the web service.
	 * @param zip The zip to query.
	 * @return The report in a string.
	 * @throws Exception
	 */
	private String generateReport(String zip) throws Exception
	{
		//establish a connection
		URL ws = new URL("http://www.innergears.com/WebServices/WeatherByZip/WeatherByZip.asmx/GetWeatherByZip?ZipCode=" + zip);
		XMLNode node = XMLParser.parse(new XmlToTxmlInputStream(ws.openStream()));
		String[] reportData = XMLUtil.selectAllValues(node, "/ArrayOfString/string");
		
		//build the report
		if(reportData.length != 23)
		{
			return "Failed to find weather data for the provided zip code.";
		}
		else
		{
			return reportData[13] + ". <br>" + 
				   reportData[1] + ", " + reportData[0] + ". <br>" +
				   reportData[7] + ". <br>" +
				   "Winds " + reportData[14] + ". <br>" +
				   "Skys " + reportData[16] + ". <br>" +
				   "Temperature " + reportData[17] + ". <br>" +
				   "Humidity " + reportData[19] + ".";
		}
	}
}
