/*
 *  Created on Oct 14, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.nukeutils;


import org.txml.*;
import java.io.*;
import java.util.*;

/**
 * The ConfManager simplifies access to an XML based configuration file. It 
 * manages file management and data location. It also provides the ability to 
 * globalize an instance itself, thus making it staticaly accessable.
 * This class acts as a wrapper for the TXML library. 
 * (http://sourceforge.net/projects/txml/)
 * 
 * @author Ryan Michela
 */
public class XmlConfManager 
{
	
	/**
	 * Reference to the global configuration file.
	 */
	private static XmlConfManager globalConf;
	
	/**
	 * The root of the XML tree.
	 */
	private XMLNode conf;
	
	/**
	 * Creates a confuguration manager and loads a file.
	 * @param confPath The path to the confuguration file to load.
	 * @throws IOException conf file not loaded.
	 */
	public XmlConfManager(String confPath) throws IOException
	{
		this(confPath, false);
	}
	
	/**
	 * Creates a confuguration manager and loads a file. Optionaly assigns the
	 * global config.
	 * @param confPath The path to the confuguration file to load.
	 * @param globalize Set to TRUE if you want this configuration file to be 
	 * the global config.
	 * @throws IOException conf file not loaded.
	 */
	public XmlConfManager(String confPath, boolean globalize) throws IOException
	{
		loadConfFile(confPath);
		if(globalize) globalConf = this;
	}
	
	/**
	 * Creates a confuguration manager and loads a file.
	 * @param confFile The confuguration file to load.
	 * @throws IOException conf file not loaded.
	 */
	public XmlConfManager(File confFile) throws IOException
	{
		this(confFile, false);
	}
	
	/**
	 * Creates a confuguration manager and loads a file. Optionaly assigns the
	 * global config.
	 * @param confFile The confuguration file to load.
	 * @param globalize Set to TRUE if you want this configuration file to be 
	 * the global config.
	 * @throws IOException conf file not loaded.
	 */
	public XmlConfManager(File confFile, boolean globalize) throws IOException
	{
		loadConfFile(confFile.getPath());
		if(globalize) globalConf = this;
	}
	
	/**
	 * Internal constructor for creating conf managers out of xml subtrees.
	 * @param conf The subtree to use.
	 */
	XmlConfManager(XMLNode conf)
	{
		this.conf = conf;
	}
	
	/**
	 * Loads and parses the configuration file.
	 * @param confPath The file to load.
	 * @throws IOException conf file not loaded.
	 */
	private void loadConfFile(String confPath) throws IOException
	{	
		try
		{
			FileInputStream fis = new FileInputStream(confPath);
			conf = XMLParser.parse(new XmlToTxmlInputStream(fis));
		}
		catch(FileNotFoundException e)
		{
			throw new IOException("Unable to access configuration file!");
		}
		catch(IOException e)
		{
			throw new IOException("Error loading configuration file!");
		}
		catch(XMLSyntaxError e)
		{
			throw new IOException("Malformed config file! " + e.getMessage());
		}
	}

	/**
	 * Loads a single string from the configuration file. If the provided
	 * confPath returns multiple matches, only the first value is returned.
	 * @param confPath The path in the configuration hiearchy to load. 
	 * Format: /RootNode/Node/Node/Element
	 * @return The value of the target node. NULL if ther are no matches.
	 * @throws IOException conf file not loaded.
	 */
	public String getOneConfString(String confPath) throws IOException
	{
		if(conf != null)
		{	
			String ret = XMLUtil.selectOneValue(conf, confPath);
			if(ret != null)
			{
				return ret;
			}
			else
			{
				return new String("");
			}
		}
		else
		{
			throw new IOException("Configuration file not loaded!");
		}
	}
	
	/**
	 * Loads multiple string from the configuration file. If the provided
	 * confPath returns multiple matches, all values are returned.
	 * @param confPath The path in the configuration hiearchy to load. 
	 * Format: /RootNode/Node/Node/Element
	 * @return The value of the target node. NULL if ther are no matches.
	 * @throws IOException conf file not loaded.
	 */
	public String[] getAllConfStrings(String confPath) throws IOException
	{
		if(conf != null)
		{
			return XMLUtil.selectAllValues(conf, confPath);
		}
		else
		{
			throw new IOException("Configuration file not loaded!");
		}
	}
	
	/**
	 * Returns an array of XmlConfManager objects that represent configuration
	 * subtrees with requested root nodes located in a designated location in 
	 * the configuration file. This behaves similar to getAllConfStrings, but 
	 * returns data structures rather than values.
	 * @param partialConfPath The path in the configuration hiearchy to the top
	 * of the node set you wish to retreive. This includes the name of the
	 * target nodes.
	 * @return An array of XmlConfManagers representing the selected subtrees.
	 * @throws IOException conf file not loaded.
	 */
	public XmlConfManager[] getConfSet(String partialConfPath) 
		throws IOException
	{
		if(conf != null)
		{	
			//obtain a root node for the desired subtree
			XMLNode root = conf.select(partialConfPath);
			if(root != null)
			{	
				//look through the subtree searching for desired tags
				ArrayList confs = new ArrayList();
				for(int i = 0; i < root.getNodeCount(); i++)
				{
					//add a new conf manager to the array
					confs.add(new XmlConfManager(root.getNode(i)));
				}
				Object[] os = confs.toArray();
				XmlConfManager[] xs = new XmlConfManager[os.length];
				for(int i = 0; i < os.length; i++)
				{
					xs[i] = (XmlConfManager)os[i];
				}
				return xs;
			}
			else
			{
				return new XmlConfManager[0];
			}
		}
		else
		{
			throw new IOException("Configuration file not loaded!");
		}
	}
	
	/**
	 * Promotes this XmlConfManager to global conf manager status.
	 */
	public void globalize()
	{
		globalConf = this;
	}
	
	/**
	 * Gets the global configuration manager.
	 * @return The global ConfManager.
	 */
	public static XmlConfManager getGlobalConf()
	{
		return globalConf;
	}
}