/*
 *  Created on Nov 16, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

package org.nukesoft.nukeutils;

import java.io.*;
import java.util.*;

/**
 * The purpose of this class is to allow the TXML library to read standard
 * XML 1.0 compliant files. This is accomplished by providing a filtering
 * layer between the file and TXML that converts or removes any incompatable
 * XML markup. Comments and metadata are removed, terminal tags are converted
 * into empty pairs of tags, and tag parameters are promoted to full fledged
 * sub-tags.
 * 
 * @author Ryan Michela <a href="http://www.nukesoft.org">(http://www.nukesoft.org)</a>
 */
public class XmlToTxmlInputStream extends InputStream
{
	private ByteArrayInputStream buf;
	
	public XmlToTxmlInputStream(InputStream in) throws IOException
	{
		//the completed buffer
		StringBuffer comp = new StringBuffer();
		
		boolean inTag       = false;
		boolean inMetadata  = false;
		boolean inAttrib    = false;
		boolean inSQuot     = false;
		boolean inDQuot     = false;
		StringBuffer tagName     = new StringBuffer();
		StringBuffer attribName  = new StringBuffer();
		StringBuffer quotStr     = new StringBuffer();
		LinkedList attribQueue = new LinkedList();
		int commentSkipCount = 0;
		int ch;
		
		//read the input stream character by character, constructing the
		//output stream
		while((ch = in.read()) != -1)
		{
			//in a tag
			if(inTag)
			{
				//in metadata
				if(inMetadata)
				{
					//read the end of metadata, possibly
					if(ch == '>')
					{
						if(commentSkipCount == 0)
						{	
							inMetadata = false;
							inTag = false;
						}
						else
						{
							commentSkipCount--;
						}
					}
					//read a commented out tag
					if(ch == '<')
					{
						commentSkipCount++;
					}
					else
					{
						//ignore character
					}
				}
				//in a tag attribute name
				else if(inAttrib)
				{
					//read the end of the attribute name
					if(ch == '=')
					{
						inAttrib = false;
					}
					//read a terminal tag with a space before the terminal slash
					//handle as a terminal tag and escape attribute mode
					else if(ch == '/')
					{
						//convert the terminal tag to a pair of tags with no value
						String tmp = "<" + tagName.toString().trim() + "></" + tagName.toString().trim() + ">";
						//append and flush
						comp.append(tmp);
						tagName = new StringBuffer();
						
						inAttrib = false;
					}
					//read a character in the attribute name
					else
					{
						attribName.append((char)ch);
					}
				}
				//in a quoted string
				else if(inSQuot || inDQuot)
				{
					//read end of quoted string
					if((ch == '"' && inDQuot) || (ch == '\'' && inSQuot))
					{
						inSQuot = false;
						inDQuot = false;
						//queue the attribute and flush the holding buffers
						String tmp = "<" + attribName.toString().trim() + ">" + quotStr.toString() + "</" + attribName.toString().trim() + ">";
						attribQueue.addLast(tmp);
						//and flush the holding buffers
						attribName = new StringBuffer();
						quotStr = new StringBuffer();
					}
					//read a character in the string body
					else
					{
						quotStr.append((char)ch);
					}
				}
				//not in metadata, nor an attrubute name, not a quoted string
				else
				{
					//read the opening of metadata or a comment
					if(ch == '?' || ch == '!')
					{
						inMetadata = true;
					}
					//read the opening of a double quoted string
					else if(ch == '"')
					{
						inDQuot = true;
					}
					//read the opening of a single quoted string
					else if(ch == '"')
					{
						inSQuot = true;
					}
					//read the start of a tag attribute
					else if(ch == ' ')
					{
						inAttrib = true;
					}
					
					//read the end of a tag
					else if(ch == '>')
					{	
						//see if tagName has any data. No data implies that 
						//either metadata or a terminal tag was read
						if(tagName.length() != 0)
						{
							//write the opening tag
							comp.append("<" + tagName.toString().trim() + ">");
							//write any queued attributes
							while(attribQueue.size() != 0)
							{
								String tmp = (String)attribQueue.removeFirst();
								comp.append(tmp);
							}
							//flush
							tagName = new StringBuffer();
							attribQueue.clear();
						}
						//we have reached the end of a tag
						inTag = false;
					}
					//read a terminal tag or a closing tag
					else if(ch == '/')
					{
						//see if tagName has any data. No data implies this /
						//comes before the tag name and thus this is a closing tag
						if(tagName.length() == 0)
						{
							tagName.append('/');
						}
						//this must be a terminal tag
						else
						{	
							//convert the terminal tag to a pair of tags with no value
							//and insert any attributes
							comp.append("<" + tagName.toString().trim() + ">");
							
							while(attribQueue.size() != 0)
							{
								String tmp = (String)attribQueue.removeFirst();
								comp.append(tmp);
							}
							
							comp.append("</" + tagName.toString().trim() + ">");
							
							//flush
							attribQueue.clear();
							tagName = new StringBuffer();
						}
					}
					//read a character in the tag name
					else
					{
						tagName.append((char)ch);
					}
				}
			}
			//not in a tag
			else
			{
				//read the opening of a tag
				if(ch == '<')
				{
					inTag = true;
				}
				//in the value of a tag
				else
				{
					comp.append((char)ch);
				}
			}
		}
		
		//now that the input has been transformed, store it in the local input stream
		//System.out.println(comp);
		
		buf = new ByteArrayInputStream(comp.toString().getBytes());
		in.close();
	}
	
	public int available() throws IOException
	{
		return buf.available();
	}
	
	public void close() throws IOException
	{
		buf.close();
	}

	public synchronized void mark(int readlimit)
	{
		buf.mark(readlimit);
	}

	public boolean markSupported()
	{
		return buf.markSupported();
	}

	public int read() throws IOException
	{
		return buf.read();
	}

	public int read(byte[] b, int off, int len) throws IOException
	{
		return buf.read(b, off, len);
	}

	public int read(byte[] b) throws IOException
	{
		return buf.read(b);
	}

	public synchronized void reset() throws IOException
	{
		buf.reset();
	}

	public long skip(long n) throws IOException
	{
		return buf.skip(n);
	}

}
