/*
 *  Created on Oct 29, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.bot2.lineproc;

import org.nukesoft.bot2.core.*;
import org.nukesoft.bot2.core.IM;
import org.nukesoft.nukeutils.*;

import java.util.*;

/**
 * The DynamicLoadLPContext provides an internal implementation of the
 * LineProcContext interface. It handles loading of the external LineProc
 * class, as well as context maintnance. For each conversational dialog,
 * there exists exactly one DynamicLoadLPContext. The context will reload
 * a different LineProc if asked to do so by the LineProc. This keeps
 * LineProc manipulation out of the Dialog implementation.
 * 
 * @author Ryan Michela
 */
public class DynamicLoadLPContext implements LineProcContext
{

	private LineProcContextCallback callback;
	private String lastLineProcID;
	private String currentLineProcID;
	private LineProc currentLineProc;
	
	/**
	 * Createst a DynamicLoadLPContext object and register a callback 
	 * and an initial LineProc.
	 * @param parent The callback for which this object is registered.
	 * @param defaultLibBindID The initial line proc bind ID to load.
	 */
	public DynamicLoadLPContext(String defaultLibBindID)
	{
		this.currentLineProcID = defaultLibBindID;
		this.lastLineProcID = defaultLibBindID;
	}
	
	/**
	 * Binds a local callback.
	 * @param callback The callback.
	 */
	public void bindCallback(LineProcContextCallback callback)
	{
		this.callback = callback;
	}

	/**
	 * Sends an IM.
	 * @param message The IM to send.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#sendIM(org.nukesoft.bot2.core.IM)
	 */
	public void sendIM(IM message)
	{
		callback.sendIM(message);
	}

	/**
	 * Sends a warning.
	 * @param screenName The sceen name to warn.
	 * @param anonymous If true, send an anonymous warning.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#sendWarning(java.lang.String, boolean)
	 */
	public void sendWarning(String screenName, boolean anonymous)
	{
		callback.sendWarning(screenName, anonymous);
	}
	
	/**
	 * Forewards an incomming IM to the LineProc.
	 * @param message The message to foreward.
	 */
	public void handleIM(IM message)
	{
		if(currentLineProc != null)
		{	
			currentLineProc.handleIM(message);
		}
		else
		{
			VerboseConsole.println(VerboseConsole.RED_TEXT() +
									"LineProc" +
									VerboseConsole.GRAY_TEXT() +
									": Failed to process incomming IM. No loaded line processor!", 
									VerboseConsole.CRITICAL);
		}
	}
	
	/**
	 * Forewards an incomming warning to the LineProc.
	 * @param screenName The screen name that did the warning.
	 */
	public void handleWarn(String screenName)
	{
		if(currentLineProc != null)
		{	
			currentLineProc.handleWarn(screenName);
		}
		else
		{
			VerboseConsole.println(VerboseConsole.RED_TEXT() +
									"LineProc" +
									VerboseConsole.GRAY_TEXT() +
									": Failed to process incomming warning. No loaded line processor!", 
									VerboseConsole.CRITICAL);
		}
	}

	/**
	 * Loads a new line processor.
	 * @param lineProcBindID Bind ID of the line processor to load.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#loadNewLineProc(java.lang.String)
	 */
	public boolean loadNewLineProc(String lineProcBindID)
	{
		String libPath = callback.getLineProcLibPath(lineProcBindID);
		if(libPath != null)
		{	
			AbsolutePathClassLoader apcl = new AbsolutePathClassLoader();
			try
			{
				//shut down the current line processor
				if(currentLineProc != null)
				{	
					currentLineProc.shutdown();
				}
				//load the new line processor
				currentLineProc = (LineProc)apcl.loadClass(libPath).newInstance();
				currentLineProc.bindContext(this);
				currentLineProc.init(getRemoteScreenName());
				//and archive the libpath for furure reference
				lastLineProcID = currentLineProcID;
				currentLineProcID = lineProcBindID;
				VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
										"LineProc" +
										VerboseConsole.GRAY_TEXT() +
										": Line Processor " + libPath + " loaded by " + callback.getIdentity(), 
										VerboseConsole.LOW);
				return true;
			}
			catch(Exception e)
			{
				VerboseConsole.println(VerboseConsole.RED_TEXT() +
										"LineProc" +
										VerboseConsole.GRAY_TEXT() +
										": Failed to load line processor!: " + libPath, 
										VerboseConsole.CRITICAL);
				return false;
			}
		}
		else
		{
			VerboseConsole.println(VerboseConsole.RED_TEXT() +
									"LineProc" +
									VerboseConsole.GRAY_TEXT() +
									": Could not find " + lineProcBindID + " in LineProc bindings!", 
									VerboseConsole.CRITICAL);
			return false;
		}
	}

	/**
	 * Tells a peer to load a new line processor.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#loadNewLineProcOnPeer(java.lang.String, java.lang.String)
	 */
	public boolean loadNewLineProcOnPeer(String dialogID, String lineProcBindID)
	{
		return callback.loadNewLineProcOnPeer(dialogID, lineProcBindID);
	}

	/**
	 * Returns the bind ID to the last loaded line processor.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#lastLoadedLineProc()
	 */
	public String lastLoadedLineProc()
	{
		return lastLineProcID;
	}

	/**
	 * Returns the current warning level.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#getWarningLevel()
	 */
	public String getWarningLevel()
	{
		return callback.getWarningLevel();
	}

	/**
	 * Returns a list of active conversations.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#getConversations()
	 */
	public Set getConversations()
	{
		return callback.getConversations();
	}


	/**
	 * Executes a conversation bind.
	 * @param dialogID The dialog to bind to.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#bindConversation(java.lang.String)
	 */
	public boolean bindConversation(String dialogID)
	{
		return callback.bindConversation(dialogID);
	}

	/**
	 * Executes a conversation unbind.
	 * @param dialogID The dialog to unbind from.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#unbindConversation(java.lang.String)
	 */
	public boolean unbindConversation(String dialogID)
	{
		return callback.unbindConversation(dialogID);
	}

	/**
	 * Returns the dialog identity string.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#getIdentity()
	 */
	public String getIdentity()
	{
		return callback.getIdentity();
	}
	
	/* (non-Javadoc)
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#getForeignScreenName()
	 */
	public String getRemoteScreenName()
	{
		String sn = callback.getIdentity();
		return sn.substring(sn.lastIndexOf('~')+1);
	}

	/**
	 * Shuts down the current loaded line processor.
	 */
	public void shutdown()
	{
		if(currentLineProc != null)
		{	
			VerboseConsole.println(VerboseConsole.BLUE_TEXT() +
									"LineProc" +
									VerboseConsole.GRAY_TEXT() +
									": Shutting down LineProc.", 
									VerboseConsole.LOW);
			currentLineProc.shutdown();
		}
		else
		{
			VerboseConsole.println(VerboseConsole.RED_TEXT() +
									"LineProc" +
									VerboseConsole.GRAY_TEXT() +
									": Cannot shut down LineProc. No LineProc loaded!", 
									VerboseConsole.CRITICAL);
		}
	}

	/**
	 * Returns a set of comm port bind IDs.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#getComPortBindIDs()
	 */
	public Set getComPortBindIDs()
	{
		return callback.getComPortBindIDs();
	}

	/**
	 * Returns a set of line proc bind IDs.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#getLineProcBindIDs()
	 */
	public Set getLineProcBindIDs()
	{
		return callback.getLineProcBindIDs();
	}

	/**
	 * Intitiates a conversation.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#initiateConversation(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void initiateConversation(
		String commPortBindID,
		String remoteScreenName,
		String salutation)
	{
		callback.initiateConversation(commPortBindID, remoteScreenName, salutation);
	}

	/**
	 * Interjects into another dialog.
	 * @see org.nukesoft.bot2.lineproc.LineProcContext#interject(java.lang.String, java.lang.String)
	 */
	public void interject(String dialogID, String interjection)
	{
		callback.interject(dialogID, interjection);
	}
}
