/*
 *  Created on Oct 16, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.bot2.lineproc;
import org.nukesoft.bot2.core.*;
import java.util.*;

/**
 * The LineProcContext interface defines the interface that line processors use
 * to talk to and control the bot core. It provides callback services for message
 * forewarding, as well as utility services.
 * 
 * @author Ryan Michela
 */
public interface LineProcContext
{
	/**
	 * Sends an IM.
	 * @param message The IM to send.
	 */
	public void sendIM(IM message);
	/**
	 * Sends a warning. <i>May not work on all networks.</i>
	 * @param screenName The screen name to warn.
	 * @param anonymous If TRUE, warn anonymously.
	 */
	public void sendWarning(String screenName, boolean anonymous);
	
	/**
	 * Fetches the warning level of the account to which this
	 * conversation is bound. 
	 * @return Integer string containing warning level if warning
	 * is supported on the supporting network. "N/A" if warning is 
	 * not.
	 */
	public String getWarningLevel();
	
	/**
	 * Returns the identity of the Dialog object. This is not the foreign
	 * screen name attached to this context, but rather the internal unique
	 * context identity. To get the foreign screen name, use the
	 * getForeignScreenName() method.
	 * @return Identity string in the form
	 * CommPortBindName~ForeignScreenName.
	 */
	public String getIdentity();
	
	/**
	 * Returns the foreign screen name attached to this context. That is,
	 * the screen name of the person at the other end of the line.
	 * @return The remote screen name.
	 */
	public String getRemoteScreenName();
	
	/**
	 * Fetches the set of all active conversations.
	 * @return Set of Strings with fully qualified conversation names
	 * in the format CommPortBindName~ForeignScreenName.
	 */
	public Set getConversations();
	/**
	 * Forces all trafic to and from designated dialog to also be
	 * echoed to this LineProc.
	 * @param dialogID Dialog ID to bind to in the format
	 * CommPortBindName~ForeignScreenName.
	 */
	public boolean bindConversation(String dialogID);
	/**
	 * Halts all traffic to and from designated dialog. In other words,
	 * stops message echoing. Does nothing if not bound to target dialog.
	 * @param dialogID Dialog ID to unbind from in the format
	 * CommPortBindName~ForeignScreenName.
	 */
	public boolean unbindConversation(String dialogID);
	
	/**
	 * Loads a new line processor and transfers control to it.
	 * @param lineProcBindID The bind ID of the class to load.
	 * @return TRUE if success, FALSE if otherwise.
	 */
	public boolean loadNewLineProc(String lineProcBindID);
	
	/**
	 * Loads a new line processor for a peer Dialog and transfers
	 * control to it.
	 * @param dialogID The peer to manipulate.
	 * @param lineProcBindID The line proc to load.
	 * @return TRUE if success, FALSE if otherwise.
	 */
	public boolean loadNewLineProcOnPeer(String dialogID, String lineProcBindID);
	
	/**
	 * The bind ID of the last loaded class.
	 * @return Path.
	 */
	public String lastLoadedLineProc();
	
	/**
	 * Returns a set of Strings containing the bind IDs of all
	 * comports. This information can be used for internal purposes
	 * or for initiating a conversation with an unknowing victim.
	 * @return A set of Strings.
	 */
	public Set getComPortBindIDs();
	/**
	 * Returns a set of Strings containing the bind IDs of all
	 * line procs. This information can be used for internal purposes
	 * or for rebinding the current conversation to a new line proc.
	 * @return A set of Strings.
	 */
	public Set getLineProcBindIDs();
	/**
	 * Initates a conversation with a remote screen name.
	 * @param commPortBindID The commPort to use for the dialog.
	 * @param remoteScreenName The screen name to initiate a conversation with.
	 * @param salutation The salutation to send them.
	 */
	public void initiateConversation(String commPortBindID, String remoteScreenName, String salutation);
	/**
	 * Sends an interjection into another Dialog's conversation flow.
	 * @param dialogID The name of the dialog to interject into.
	 * @param interjection The text to interject with.
	 */
	public void interject(String dialogID, String interjection);
}
