/*
 *  Created on Nov 20, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.nukeutils;

import java.util.*;
/**
 * The ExtendedStringTokenizer class us used much the same way as a StringTokenizer,
 * but provides some enhanced functionality usefull to writing command parsers.
 * These include token peeking, parsing quoted strings as single tokens, and returning 
 * the rest of the token stream as a single string.
 * 
 * @author Ryan Michela
 */
public class ExtendedStringTokenizer
{
	private StringTokenizer st;
	private String str;
	
	public ExtendedStringTokenizer(String str)
	{
		this.str = str;
		this.st = new StringTokenizer(str);
	}
	/**
	 * Calculates the number of times that this tokenizer's nextToken method 
	 * can be called before it generates an exception. This does not take
	 * quoted tokens into account. The current position is not advanced.
	 * @return The number of tokens remaining in the string.
	 */
	public int countTokens()
	{
		return st.countTokens();
	}
	
	/**
	 * Tests if there are more tokens available from this tokenizer's string. 
	 * If this method returns true, then a subsequent call tonextToken with 
	 * no argument will successfully return a token.
	 * @return TRUE if and only if there is at least one token in the string 
	 * after the current position; FALSE otherwise.
	 */
	public boolean hasMoreTokens()
	{
		return st.hasMoreTokens();
	}
	
	/**
	 * Returns the next token from this string tokenizer.
	 * @return The next token from this string tokenizer.
	 * @throws NoSuchElementException if there are no more 
	 * tokens in this tokenizer's string.
	 */
	public String nextToken() throws NoSuchElementException
	{
		return st.nextToken();
	}
	
	/**
	 * Returns the next quoted token from this string tokenizer.
	 * That is, if the next token starts a quoted string,
	 * the entire quoted string will be returned sans quote
	 * characters, otherwise, the next token will be returned.
	 * @return The next quoted token from this string tokenizer.
	 * @throws NoSuchElementException if there are no more 
	 * tokens in this tokenizer's string.
	 */
	public String nextQuotedToken() throws NoSuchElementException
	{
		String tok = st.nextToken();
		//if the next token starts a quoted string, try to find the
		//rest of the string.
		if(tok.startsWith("\""))
		{
			String quote = new String(tok);
			//check to see of the first token ends with a string
			if(!quote.endsWith("\""))
			{	
				//keep adding tokens to the string.
				//stop at end of token stream or when end quote is found
				while(st.hasMoreTokens())
				{
					tok = st.nextToken();
					quote += (" " + tok);
					if(tok.endsWith("\""))
					{
						break;
					}
				}
			}
			//trim off quote characters and return
			return quote.substring(1, quote.length() - 1);
		}
		else
		{
			return tok;
		}
	}
	
	/**
	 * Returns the next token from this string tokenizer without
	 * actually removing the token from the token stream. This method is
	 * costly so try to cache the results whenever possible.
	 * @return The next token from this string tokenizer.
	 * @throws NoSuchElementException if there are no more 
	 * tokens in this tokenizer's string.
	 */
	public String peekNextToken() throws NoSuchElementException
	{
		String next = nextToken();
		String rest = restTokens();
		st = new StringTokenizer(next + " " + rest);
		return next;
	}
	
	/**
	 * Returns all remaining tokens in a single string. If there are
	 * no more tokens remaining, an empty string is returned.
	 * @return The remaining tokens.
	 */
	public String restTokens()
	{
		if(st.hasMoreTokens())
		{
			String rest = st.nextToken();
			while(st.hasMoreTokens())
			{
				rest += (" " + st.nextToken());
			}
			return rest;
		}
		else
		{
			return "";
		}
	}
	
	/**
	 * Determines if this tokenizer's input string contains a given token.
	 * @param token The token to test.
	 * @return TRUE if the token is present, FALSE otherwise.
	 */
	public boolean containsToken(String token)
	{
		StringTokenizer stTmp = new StringTokenizer(str);
		boolean found = false;
		while(stTmp.hasMoreTokens())
		{
			if(stTmp.nextToken().equals(token))
			{
				found = true;
			}
		}
		return found;
	}
	
	/**
	 * Determines if this tokenizer's input string contains a given token.
	 * Case is ignored in durring comparison.
	 * @param token The token to test.
	 * @return TRUE if the token is present, FALSE otherwise.
	 */
	public boolean containsTokenIgnoreCase(String token)
	{
		StringTokenizer stTmp = new StringTokenizer(str);
		boolean found = false;
		while(stTmp.hasMoreTokens())
		{
			if(stTmp.nextToken().equalsIgnoreCase(token))
			{
				found = true;
			}
		}
		return found;
	}
}
