/*
 *  Created on Oct 14, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

package org.nukesoft.nukeutils;

import java.io.*;
import java.util.*;

/**
 * The Log class is used for autonimous log maintnance. One file is created for
 * each path given, regardless of the number of instances referenceing that file.
 * File maintnance is done automaticaly, as are timed log flushes and timestamping.
 * Log enteries from multiple instances are automaticaly interleved within the file.
 * @author Ryan Michela
 */
public class Log
{
	private static HashMap files = new HashMap();
	private static LogService ls = new LogService(0);
	private static String workingDirectory = "";
	private String logFileName;
	
	/**
	 * Creates a Log object. If this is the first Log object created,
	 * the log service is started.
	 * @param logFileName Path to the log file.
	 */
	public Log(String logFileName)
	{
		this.logFileName = logFileName;
		//only create a new log file if none exists for given file name
		if(!files.containsKey(logFileName))
		{
			try
			{
				files.put(logFileName, new LogFile(logFileName));
			}
			catch(IOException e)
			{
				VerboseConsole.println(VerboseConsole.RED_TEXT() +
										"WARNING!" +
										VerboseConsole.GRAY_TEXT() +
										" Unable to access log " + logFileName + " Ignoring all log enteries.", 
										VerboseConsole.CRITICAL);
				files.put(logFileName, null);
			}
		}
	}
	
	/**
	 * Creates an entry in the log.
	 * @param entry Text to add.
	 */
	public void logEntry(String entry)
	{
		LogFile lf = (LogFile)files.get(logFileName);
		if(lf != null)
		{
			lf.addEntry(entry);
		}
	}
	
	/**
	 * Manualy flushes this log file to disk.
	 */
	public void flush()
	{
		LogFile lf = (LogFile)files.get(logFileName);
		if(lf != null)
		{
			lf.flush();
		}
	}
	
	/**
	 * Closes a log file and removes it from the auto flush table.
	 */
	public void close()
	{
		this.flush();
		LogFile lf = ((LogFile) files.get(logFileName));
		if(lf != null)
		{	
			lf.flush();
			synchronized(files)
			{
				files.remove(logFileName);
			}
		}
	}
	
	/**
	 * Manualy flushes all log files to disk.
	 */
	public static void flushAll()
	{
		synchronized(files)
		{
			Iterator it = files.values().iterator();
			while(it.hasNext())
			{
				LogFile lf = (LogFile)it.next();
				if(lf != null)
				{
					lf.flush();
				}
			}
		}
	}
	
	/**
	 * Properly shuts down the log service and flushes all files.
	 * <b>Call before terminating your program!</b>
	 */
	public static void shutdown()
	{
		synchronized(files)
		{
			Iterator it = files.values().iterator();
			while(it.hasNext())
			{
				LogFile lf = (LogFile)it.next();
				if(lf != null)
				{
					lf.shutdown();
				}
			}
			ls.shutdown();
		}
	}
	
	/**
	 * Sets the time between global log flushes.
	 * @param timeBetweenFlushes The time in seconds. Zero (0) for manual flushing only.
	 */
	public static void setTimeBetweenFlushes(long timeBetweenFlushes) //in seconds
	{
		ls.setTimeBetweenFlushes(timeBetweenFlushes);
	}
	
	/**
	 * Sets the working directory for all logs.
	 * @param workingDirectory The directory path.
	 */
	public static void setWorkingDirectory(String workingDirectory)
	{
		Log.workingDirectory = workingDirectory;
	}

	/**
	 * Utility class to mantain a physical log file.
	 * 
	 * @author Ryan Michela
	 */
	private static class LogFile
	{
		private PrintWriter pw;
		private String fileName;
		
		/**
		 * Opens a file and writes the log header.
		 * @param fileName The file path to open.
		 * @throws IOException
		 */
		public LogFile(String fileName) throws IOException
		{
			this.fileName = fileName;
			
			String wd = Log.workingDirectory;
			if(!wd.endsWith("/") && !wd.endsWith("\\"))
			{
				wd += "/";
			}
			
			pw = new PrintWriter(new BufferedWriter(new FileWriter(wd + fileName, true)));
			addEntry("========Opening Log========");
		}
		
		/**
		 * Adds a single entry to the log.
		 * @param logString The string to log.
		 */
		public void addEntry(String logString)
		{
			pw.println(timestamp() + " :: " + logString);
		}
		
		/**
		 * Returns the path this log is working on.
		 * @return Path to log file.
		 */
		public String getFileName()
		{
			return fileName;
		}
		
		/**
		 * Flushes this individual log file.
		 */
		public void flush()
		{
			addEntry("Log Flushed");
			pw.flush();
		}
		
		/**
		 * Generates a timestamp.
		 * @return String containing the time.
		 */
		private String timestamp()
		{
			Calendar now = Calendar.getInstance();
			return (now.get(Calendar.DAY_OF_MONTH) > 9 ? ("" + now.get(Calendar.DAY_OF_MONTH)) : ("0" + now.get(Calendar.DAY_OF_MONTH))) + "/" +
			       (now.get(Calendar.MONTH) > 9 ? ("" + now.get(Calendar.MONTH)) : ("0" + now.get(Calendar.MONTH))) + "/" +
			       now.get(Calendar.YEAR) + " " +
			       (now.get(Calendar.HOUR_OF_DAY) > 9 ? ("" + now.get(Calendar.HOUR_OF_DAY)) : ("0" + now.get(Calendar.HOUR_OF_DAY))) + ":" +
			       (now.get(Calendar.MINUTE) > 9 ? ("" + now.get(Calendar.MINUTE)) : ("0" + now.get(Calendar.MINUTE))) + ":" +
			       (now.get(Calendar.SECOND) > 9 ? ("" + now.get(Calendar.SECOND)) : ("0" + now.get(Calendar.SECOND)));
		}
		
		/**
		 * Closes this log file.
		 */
		public void shutdown()
		{
			addEntry("========Closing Log========");
			pw.close();
		}
	}
	
	/**
	 * Utility class that autoflushes all logs.
	 * 
	 * @author Ryan Michela
	 */
	private static class LogService extends Thread
	{
		private long timeBetweenFlushes;
		private boolean keepGoing = true;
		private boolean pause = false;
		
		/**
		 * Starts the log service thread.
		 * @param timeBetweenFlushes The time, in seconds, between flushes.
		 */
		public LogService(long timeBetweenFlushes)
		{
			super("Log Service");
			this.timeBetweenFlushes = timeBetweenFlushes;
			this.setPriority(Thread.MIN_PRIORITY);
			
			if(timeBetweenFlushes <= 0)
			{
				pause = true;
			}
			this.start();
		}
		
		/**
		 * Sets the flush interval.
		 * @param timeBetweenFlushes Flush interval, in seconds.
		 */
		public void setTimeBetweenFlushes(long timeBetweenFlushes) //in seconds
		{
			this.timeBetweenFlushes = timeBetweenFlushes;
			if(timeBetweenFlushes <= 0)
			{
				pause = true;
			}
			else
			{
				pause = false;
				synchronized(this)
				{
					notify();
				}
			}
			
		}
		
		/**
		 * Terminates the log service.
		 */
		public void shutdown()
		{
			keepGoing = false;
			this.interrupt();
		}
		
		/** Main log service thread loop.
		 * @see java.lang.Runnable#run()
		 */
		public void run()
		{
			VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
									"Starting Log Service..." +
									VerboseConsole.GRAY_TEXT(), 
									VerboseConsole.HIGH);
			while(keepGoing)
			{
				try
				{
					if(pause)
					{
						synchronized(this)
						{
							wait();
						}
					}
					sleep(timeBetweenFlushes * 1000);
				}
				catch(InterruptedException e)
				{
					
				}
				VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
										"Logging Service" +
										VerboseConsole.GRAY_TEXT() +
										": Logs flushed", 
										VerboseConsole.LOW);
				Log.flushAll();
			}
			Log.flushAll();
			VerboseConsole.println(VerboseConsole.GREEN_TEXT() +
									"Stopping Log Service..." +
									VerboseConsole.GRAY_TEXT(), 
									VerboseConsole.HIGH);
		}
	}
}
