/*
 *  Created on Oct 14, 2003
 *  Copyright 2003 by Ryan Michela
 *
 *   This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package org.nukesoft.nukeutils;

/**
 * The VerboseConsole class is used to dynamicaly define the verbosity of all
 * console output. This class can be used as a substitute to System.out. Class
 * constants can be used to implement ANSI colors and text effects.
 * 
 * Example: <code>VerboseConsole.println(RED_BACK() +
 * 										 BLINK_ON() +
 * 										 "This is blinking text" +
 * 										 BLINK_OFF() +
 * 										 " This is not" +
 * 										 RESET_COLORS());</code>
 * 
 * @author Ryan Michela
 */
public class VerboseConsole 
{
	//verbosity levels
	public static final int SILENT   = 5;
	public static final int LOW      = 1;
	public static final int MEDIUM   = 2;
	public static final int HIGH     = 3;
	public static final int CRITICAL = 4;

	private static final int DEFAULT_LEVEL = LOW;
	private static int currentLevel;
	private static boolean ANSIon = false;

	public VerboseConsole() 
	{
		this(DEFAULT_LEVEL);
	}
	
	public VerboseConsole(int level)
	{
		setLevel(level);
	}
	
	/**
	 * Sets the verbocity level of the console. Only output with a level
	 * equal to or higher to the target level will be displayed.
	 * @param level The target verbocity level.
	 */
	public static void setLevel(int level)
	{
		currentLevel = level;
	}
	
	/**
	 * Enables or disables ANSI color. ANSI is off by default.
	 * @param isOn TRUE = on, FALSE = off.
	 */
	public static void enableANSI(boolean isOn)
	{
		ANSIon = isOn;
		if(isOn)
		{
			VerboseConsole.println("Console in " + 
								   VerboseConsole.BOLD_ON() + 
								   VerboseConsole.RED_TEXT() + 
								   "C" +
								   VerboseConsole.BROWN_TEXT() +
								   "O" +
								   VerboseConsole.GREEN_TEXT() + 
								   "L" +
								   VerboseConsole.CYAN_TEXT() + 
								   "O" +
								   VerboseConsole.BLUE_TEXT() + 
								   "R" + 
								   VerboseConsole.RESET_COLORS(),
								   VerboseConsole.LOW);
		}
	}
	
	/**
	 * Print a string at a given level. Equivilent to System.out.print();
	 * @param s The string to print.
	 * @param level The level of this message.
	 */
	public static void print(String s, int level)
	{
		if(level >= currentLevel)
		{
			System.out.print(s);
		}
	}
	
	/**
	 * Print a string at the default level. Equivilent to System.out.print();
	 * @param s The string to print.
	 */
	public static void print(String s)
	{
		print(s, DEFAULT_LEVEL);
	}
	
	/**
	 * Print a string at a given level. Equivilent to System.out.println();
	 * @param s The string to print.
	 * @param level The level of this message.
	 */
	public static void println(String s, int level)
	{
		if(level >= currentLevel)
		{
			System.out.println(s);
		}
	}
	
	/**
	 * Print a string at the default level. Equivilent to System.out.println();
	 * @param s The string to print.
	 */
	public static void println(String s)
	{
		println(s, DEFAULT_LEVEL);
	}
	
	
	
	//	foreground color
	public static final String BLACK_TEXT() { return ANSIon ? "\033[30m" : "";}
	public static final String RED_TEXT() { return ANSIon ? "\033[31m" : "";}
	public static final String GREEN_TEXT() { return ANSIon ? "\033[32m" : "";}
	public static final String BROWN_TEXT() { return ANSIon ? "\033[33m" : "";}
	public static final String BLUE_TEXT() { return ANSIon ? "\033[34m" : "";}
	public static final String MAGENTA_TEXT() { return ANSIon ? "\033[35m" : "";}
	public static final String CYAN_TEXT() { return ANSIon ? "\033[36m" : "";}
	public static final String GRAY_TEXT() { return ANSIon ? "\033[37m" : "";}
	
	//background color
	public static final String BLACK_BACK() { return ANSIon ? "\033[40m" : "";}
	public static final String RED_BACK() { return ANSIon ? "\033[41m" : "";}
	public static final String GREEN_BACK() { return ANSIon ? "\033[42m" : "";}
	public static final String BROWN_BACK() { return ANSIon ? "\033[43m" : "";}
	public static final String BLUE_BACK() { return ANSIon ? "\033[44m" : "";}
	public static final String MAGENTA_BACK() { return ANSIon ? "\033[45m" : "";}
	public static final String CYAN_BACK() { return ANSIon ? "\033[46m" : "";}
	public static final String WHITE_BACK() { return ANSIon ? "\033[47m" : "";}
	
	//ANSI control chars
	public static final String RESET_COLORS() { return ANSIon ? "\033[0m" : "";}
	public static final String BOLD_ON() { return ANSIon ? "\033[1m" : "";}
	public static final String BLINK_ON() { return ANSIon ? "\033[5m" : "";}
	public static final String REVERSE_ON() { return ANSIon ? "\033[7m" : "";}
	public static final String BOLD_OFF() { return ANSIon ? "\033[22m" : "";}
	public static final String BLINK_OFF() { return ANSIon ? "\033[25m" : "";}
	public static final String REVERSE_OFF() { return ANSIon ? "\033[27m" : "";}
	
	
	/**
	 * Does a demo of what ANSI color can do.
	 * @param args Optionaly, the verbosity level to use.
	 */
	public static void main(String args[])
	{
		VerboseConsole.enableANSI(true);
		//set verbosity level
		if(args.length >= 1)
		{
			if(Integer.parseInt(args[0]) >= VerboseConsole.LOW && 
			   Integer.parseInt(args[0]) <= VerboseConsole.SILENT)
			{
				VerboseConsole.setLevel(Integer.parseInt(args[0]));
			}
			else if(Integer.parseInt(args[0]) == 0)
			{
				VerboseConsole.println("ANSI disabled.");
				VerboseConsole.enableANSI(false);
			}
			else
			{
				VerboseConsole.println(VerboseConsole.RED_TEXT() + 
									  "WARNING! " +
									   VerboseConsole.RESET_COLORS() + 
								      "Invalid verbosity level. Using default.");
			}
		}
		else
		{
			VerboseConsole.println( VerboseConsole.BLUE_TEXT() + 
								   "Note! " +
								   VerboseConsole.RESET_COLORS() + 
								   "Using default verbosity level.");
		}
		
		//do the demo
		VerboseConsole.println(VerboseConsole.BLINK_ON() + 
							   VerboseConsole.BOLD_ON() + 
							   VerboseConsole.RED_TEXT() + 
							   "WARNING! " + 
							   VerboseConsole.BLINK_OFF() + 
							   VerboseConsole.BOLD_OFF() + 
							   "This is a critical message." + 
							   VerboseConsole.RESET_COLORS(),
							   VerboseConsole.CRITICAL);
		
		VerboseConsole.println(VerboseConsole.BLUE_BACK() + 
							   VerboseConsole.MAGENTA_TEXT() + 
							   "Fun with colors. " + 
							   VerboseConsole.REVERSE_ON() + 
							   "Fun with reverse video." + 
							   VerboseConsole.RESET_COLORS());
		
		VerboseConsole.println("Text can have many " + 
							   VerboseConsole.BOLD_ON() + 
							   VerboseConsole.RED_TEXT() + 
							   "C" +
							   VerboseConsole.BROWN_TEXT() +
							   "O" +
							   VerboseConsole.GREEN_TEXT() + 
							   "L" +
							   VerboseConsole.CYAN_TEXT() + 
							   "O" +
							   VerboseConsole.BLUE_TEXT() + 
							   "R" +
							   VerboseConsole.MAGENTA_TEXT() + 
							   "S" + 
							   VerboseConsole.RESET_COLORS(),
							   VerboseConsole.LOW);
	}
}
